{
  "cells": [
    {
      "cell_type": "markdown",
      "metadata": {
        "id": "gHoQThpObPSq"
      },
      "source": [
        "# 第五次作业\n",
        "\n",
        "在本次作业中，大家需要完成GraphSAGE模型的训练，重点体会Neighbor Sampling的方式及用法，并在cora数据集上进行测试。\n",
        "\n",
        "本作业需要安装[CogDL](https://github.com/THUDM/cogdl)：pip install cogdl\n",
        "\n",
        "如需使用gpu版，请先安装gpu版本的[PyTorch](https://pytorch.org/get-started/locally/)，再安装cogdl。\n",
        "\n",
        "本作业由智谱GNN中心及课程团队筹备，由CogDL团队提供技术支持。\n"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "### 1. GraphSAGE模型\n",
        "仔细阅读GraphSAGE模型的实现，特别是forward/inference函数。"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 1,
      "metadata": {},
      "outputs": [],
      "source": [
        "import os\n",
        "import torch\n",
        "import torch.nn as nn\n",
        "import torch.nn.functional as F\n",
        "\n",
        "from cogdl.layers import SAGELayer\n",
        "\n",
        "\n",
        "class GraphSAGE(nn.Module):\n",
        "\n",
        "    def __init__(self, num_features, num_classes, hidden_size, num_layers, dropout, aggr=\"mean\"):\n",
        "        super(GraphSAGE, self).__init__()\n",
        "        self.adjlist = {}\n",
        "        self.num_features = num_features\n",
        "        self.num_classes = num_classes\n",
        "        self.hidden_size = hidden_size\n",
        "        self.num_layers = num_layers\n",
        "        self.dropout = dropout\n",
        "        shapes = [num_features] + [hidden_size] * (num_layers - 1) + [num_classes]\n",
        "        self.convs = nn.ModuleList(\n",
        "            [SAGELayer(shapes[layer], shapes[layer + 1], aggr=aggr) for layer in range(num_layers)]\n",
        "        )\n",
        "\n",
        "    def forward(self, x, adjs):\n",
        "        device = next(self.parameters()).device\n",
        "        for i, (src_id, graph, size) in enumerate(adjs):\n",
        "            graph = graph.to(device)\n",
        "            output = self.convs[i](graph, x)\n",
        "            x = output[: size[1]]\n",
        "            if i != self.num_layers - 1:\n",
        "                x = F.relu(x)\n",
        "                x = F.dropout(x, p=self.dropout, training=self.training)\n",
        "        return x\n",
        "\n",
        "    def inference(self, x_all, data_loader):\n",
        "        device = next(self.parameters()).device\n",
        "        for i in range(len(self.convs)):\n",
        "            output = []\n",
        "            for src_id, graph, size in data_loader:\n",
        "                x = x_all[src_id].to(device)\n",
        "                graph = graph.to(device)\n",
        "                x = self.convs[i](graph, x)\n",
        "                x = x[: size[1]]\n",
        "                if i != self.num_layers - 1:\n",
        "                    x = F.relu(x)\n",
        "                output.append(x.cpu())\n",
        "            x_all = torch.cat(output, dim=0)\n",
        "        return x_all\n"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {
        "id": "OAPbTveYjONP"
      },
      "source": [
        "### 2. 从cogdl中加载cora数据集（x表示特征，y表示标签，mask表示训练/验证/测试集的划分）"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 2,
      "metadata": {
        "colab": {
          "base_uri": "https://localhost:8080/"
        },
        "id": "61aEVO2ySq27",
        "outputId": "6da332b4-d5aa-4baf-d95e-4dd5b4b291da"
      },
      "outputs": [
        {
          "name": "stdout",
          "output_type": "stream",
          "text": [
            "Graph(x=[2708, 1433], y=[2708], train_mask=[2708], val_mask=[2708], test_mask=[2708], edge_index=[2, 10556])\n"
          ]
        }
      ],
      "source": [
        "from cogdl.datasets import build_dataset_from_name\n",
        "\n",
        "dataset = build_dataset_from_name(\"cora\")\n",
        "data = dataset[0]\n",
        "print(data)"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {
        "id": "FcRyHUuIjhfT"
      },
      "source": [
        "### 3. 使用GraphSAGE模型进行训练"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "#### 3.1 测试CogDL提供的NeighborSampler"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 3,
      "metadata": {},
      "outputs": [
        {
          "name": "stdout",
          "output_type": "stream",
          "text": [
            "tensor([0, 1, 2, 3, 4, 5, 6, 7])\n",
            "tensor([   0,    1,    2,    3,    4,    5,    6,    7, 1862,  633,  652, 1454,\n",
            "        2544, 2175, 1016, 1659, 1629, 1042,  373,  208, 1986,  332, 1761, 1416,\n",
            "         926, 1701,  470, 2176,  982,  561, 1711, 1517, 1047, 1025])\n",
            "[(tensor([   0,    1,    2,    3,    4,    5,    6,    7, 1862,  633,  652, 1454,\n",
            "        2544, 2175, 1016, 1659, 1629, 1042,  373,  208, 1986,  332, 1761, 1416,\n",
            "         926, 1701,  470, 2176,  982,  561, 1711, 1517, 1047, 1025]), Graph(col=[35]), (34, 20)), (tensor([   0,    1,    2,    3,    4,    5,    6,    7, 1862,  633,  652, 1454,\n",
            "        2544, 2175, 1016, 1659, 1629, 1042,  373,  208]), Graph(col=[14]), (20, 8))]\n"
          ]
        }
      ],
      "source": [
        "from cogdl.data.sampler import NeighborSampler\n",
        "\n",
        "data_loader = NeighborSampler(\n",
        "    dataset=dataset,\n",
        "    mask=dataset.data.train_mask,\n",
        "    sizes=[2, 2],\n",
        "    shuffle=True,\n",
        "    batch_size=8,\n",
        ")\n",
        "\n",
        "###################\n",
        "##### 作业填空 #####\n",
        "###################\n",
        "# 阅读sampler源码 https://github.com/THUDM/cogdl/blob/master/cogdl/data/sampler.py\n",
        "# 使用自己的语言来描述NeighborSampler返回的每个batch中各个元素的形式及含义\n",
        "# target_id: \n",
        "# n_id: \n",
        "# adjs: \n",
        "\n",
        "for batch in data_loader:\n",
        "    target_id, n_id, adjs = batch\n",
        "    print(target_id)\n",
        "    print(n_id)\n",
        "    print(adjs)\n",
        "    break"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "#### 3.2 使用NeighborSampler来训练GraphSAGE模型"
      ]
    },
    {
      "cell_type": "code",
      "execution_count": 4,
      "metadata": {
        "colab": {
          "base_uri": "https://localhost:8080/"
        },
        "id": "pr5lZC2eWCYm",
        "outputId": "c42b59ef-9637-4bac-e1f1-79a3f9f6228a"
      },
      "outputs": [
        {
          "name": "stderr",
          "output_type": "stream",
          "text": [
            "Epoch: 099, Train Loss: 0.0003, Val Loss: 0.8542, Val Acc: 0.7740: 100%|██████████| 100/100 [00:23<00:00,  4.31it/s]\n"
          ]
        },
        {
          "name": "stdout",
          "output_type": "stream",
          "text": [
            "Val Acc 0.79\n",
            "Test Acc 0.817\n"
          ]
        }
      ],
      "source": [
        "import math\n",
        "import copy\n",
        "import numpy as np\n",
        "from tqdm import tqdm\n",
        "from cogdl.data.sampler import NeighborSampler\n",
        "\n",
        "\n",
        "def accuracy(y_pred, y_true):\n",
        "    y_true = y_true.squeeze().long()\n",
        "    preds = y_pred.max(1)[1].type_as(y_true)\n",
        "    correct = preds.eq(y_true).double()\n",
        "    correct = correct.sum().item()\n",
        "    return correct / len(y_true)\n",
        "\n",
        "def train(hidden_size=64, batch_size=32, sizes=[10, 10]):\n",
        "    hidden_size = hidden_size\n",
        "    model = GraphSAGE(data.x.shape[1], data.y.max() + 1, hidden_size, num_layers=2, dropout=0.5)\n",
        "\n",
        "    train_loader = NeighborSampler(\n",
        "        dataset=dataset,\n",
        "        mask=dataset.data.train_mask,\n",
        "        sizes=[10, 10],\n",
        "        shuffle=True,\n",
        "        batch_size=batch_size,\n",
        "    )\n",
        "    val_loader = NeighborSampler(\n",
        "        dataset=dataset,\n",
        "        mask=dataset.data.val_mask,\n",
        "        sizes=[10, 10],\n",
        "        shuffle=False,\n",
        "        batch_size=batch_size,\n",
        "    )\n",
        "    test_loader = NeighborSampler(\n",
        "        dataset=dataset,\n",
        "        mask=None,\n",
        "        sizes=[-1],\n",
        "        shuffle=False,\n",
        "        batch_size=batch_size,\n",
        "    )\n",
        "\n",
        "    optimizer = torch.optim.Adam(model.parameters(), lr=0.01)\n",
        "    epoch_iter = tqdm(range(100), position=0, leave=True)\n",
        "    best_model = None\n",
        "    best_loss = 1e8\n",
        "    for epoch in epoch_iter:\n",
        "        model.train()\n",
        "        optimizer.zero_grad()\n",
        "        train_loss = []\n",
        "        for batch in train_loader:\n",
        "            target_id, n_id, adjs = batch\n",
        "            ###################\n",
        "            ##### 作业填空 #####\n",
        "            ###################\n",
        "            # 将每个batch的数据整理成模型forward需要的形式（可能需要使用data.x）\n",
        "            logits = model(...)\n",
        "\n",
        "            y = data.y[target_id]\n",
        "            loss = F.cross_entropy(logits, y)\n",
        "            loss.backward()\n",
        "            optimizer.step()\n",
        "            train_loss.append(loss.item())\n",
        "        train_loss = np.mean(train_loss)\n",
        "\n",
        "        model.eval()\n",
        "        val_loss = []\n",
        "        val_acc = []\n",
        "        with torch.no_grad():\n",
        "            val_count = 0\n",
        "            for batch in val_loader:\n",
        "                target_id, n_id, adjs = batch\n",
        "                ###################\n",
        "                ##### 作业填空 #####\n",
        "                ###################\n",
        "                # 将每个batch的数据整理成模型forward需要的形式（与前面相同）\n",
        "                logits = model(...)\n",
        "\n",
        "                y = data.y[target_id]\n",
        "                val_loss.append(F.cross_entropy(logits, y).item() * y.shape[0])\n",
        "                val_acc.append(accuracy(logits, y) * y.shape[0])\n",
        "                val_count += y.shape[0]\n",
        "\n",
        "            val_loss = np.sum(val_loss) / val_count\n",
        "            val_acc = np.sum(val_acc) / val_count\n",
        "            if val_loss < best_loss:\n",
        "                best_loss = val_loss\n",
        "                best_val_acc = val_acc\n",
        "                best_model = copy.deepcopy(model)\n",
        "\n",
        "        epoch_iter.set_description(f\"Epoch: {epoch:03d}, Train Loss: {train_loss:.4f}, Val Loss: {val_loss:.4f}, Val Acc: {val_acc:.4f}\")\n",
        "\n",
        "    with torch.no_grad():\n",
        "        logits = best_model.inference(data.x, test_loader)\n",
        "        val_acc = accuracy(logits[data.val_mask], data.y[data.val_mask])\n",
        "        test_acc = accuracy(logits[data.test_mask], data.y[data.test_mask])\n",
        "    print(\"Val Acc\", val_acc)\n",
        "    print(\"Test Acc\", test_acc)\n",
        "\n",
        "train()"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": [
        "### 4. 实验总结\n",
        "基于Neighbor Sampling的GraphSAGE模型训练与之前的GCN模型有何不同？GraphSAGE模型中的inference与forward有什么不同？\n"
      ]
    },
    {
      "cell_type": "markdown",
      "metadata": {},
      "source": []
    }
  ],
  "metadata": {
    "accelerator": "GPU",
    "colab": {
      "collapsed_sections": [],
      "name": "write-your-first-gcn",
      "provenance": []
    },
    "kernelspec": {
      "display_name": "Python 3.6.15 ('py3.6')",
      "language": "python",
      "name": "python3"
    },
    "language_info": {
      "codemirror_mode": {
        "name": "ipython",
        "version": 3
      },
      "file_extension": ".py",
      "mimetype": "text/x-python",
      "name": "python",
      "nbconvert_exporter": "python",
      "pygments_lexer": "ipython3",
      "version": "3.6.15"
    },
    "vscode": {
      "interpreter": {
        "hash": "7703d8cc28a94efd96474faf641471b2ac55415e3c97cc8df6cd7b51cf78a952"
      }
    }
  },
  "nbformat": 4,
  "nbformat_minor": 0
}
